package org.nextframework.persistence;

import java.util.List;

import org.nextframework.controller.crud.FiltroListagem;

public interface DAO<BEAN> {

	/**
	 * Salva determinado bean.
	 * Se quiser alterar o funcionamento do SaveOrUpdateStrategy sobrescreva o mtodo updateSaveOrUpdate.
	 * Se o bean j possuir um id, ser feito um UPDATE. Se no ser feito um INSERT.
	 * Caso tenha sido feito um INSERT, aps esse mtodo retornar o bean passado como parmetro,
	 * possuir um id. 
	 * @param bean O objeto a ser salvo.
	 */
	public abstract void saveOrUpdate(BEAN bean);

	/**
	 * Carrega determinado bean. Esse bean dever ter algum valor no ID. 
	 * O retorno ser um bean com as propriedades carregadas.
	 * @param bean
	 * @return
	 */
	public abstract BEAN load(BEAN bean);

	/**
	 * Carrega o bean para a tela de entrada de dados, no caso de um CRUD.
	 * Se quiser atualizar o QueryBuilder para essa query, sobrescreva o mtodo updateEntradaQuery
	 * @param bean
	 * @return
	 */
	public abstract BEAN loadForEntrada(BEAN bean);

	/**
	 * Retorna todos os objetos encontrados no banco de dados.
	 * Pode ser informado quais campos devem ser carregados.
	 * Esse mtodo  utilizado pelos combo boxes nos JSPs.
	 * @param extraFields
	 * @return
	 */
	public abstract List<BEAN> findForCombo(String... extraFields);

	/**
	 * Executa um find utilizando como filtro o objeto passado. <BR>
	 * Por exemplo, supondo que estamos em um DAO de municipio, e passamos como parmetro um objeto do tipo estado.
	 * Ser procurado no bean municipio qual campo da classe faz referencia ao estado, e esse campo ser utilizado como filtro.
	 * O parametro do filtro ser o objeto passado para esse mtodo.
	 * @see findBy(Object o, boolean forCombo, String... extraFields)
	 * @param o Objeto que deve ser utilizado como filtro.
	 * @return
	 */
	public abstract List<BEAN> findBy(Object o);

	/**
	 * Executa um find utilizando como filtro o objeto passado. <BR>
	 * Por exemplo, supondo que estamos em um DAO de municipio, e passamos como parmetro um objeto do tipo estado.
	 * Ser procurado no bean municipio qual campo da classe faz referencia ao estado, e esse campo ser utilizado como filtro.
	 * O parametro do filtro ser o objeto passado para esse mtodo.
	 * @param o Objeto que deve ser utilizado como filtro.
	 * @param forCombo indica se  para carregar somente o id e o descriptionProperty (Utilizado em combo boxes nos JSPs)
	 * @param extraFields Campos extras que devem ser carregados
	 * @return
	 */
	public abstract List<BEAN> findBy(Object o, boolean forCombo, String... extraFields);

	/**
	 * Retorna uma lista com todos os beans encontrados no banco, ordenados pela anotao @OrderBy da classe
	 * @return
	 */
	public abstract List<BEAN> findAll();

	/**
	 * Retorna uma lista com todos os beans encontrados no banco 
	 * ordenados por determinada propriedade
	 * @param orderBy propriedade que deve ser utilizada na ordenao
	 * @return
	 */
	public abstract List<BEAN> findAll(String orderBy);

	/**
	 * Executa um find para a tela de listagem de dados, utilizando determinado filtro.
	 * Para atualizar a query sobrescreva o mtodo updateListagemQuery.
	 * @param filtro
	 * @return
	 */
	public abstract ResultList<BEAN> findForListagem(FiltroListagem filtro);

	/**
	 * Exclui determinado bean do banco de dados.
	 * Utilizar caso exista a transao atual.
	 * @param bean
	 */
	public abstract void delete(final BEAN bean);

}